/*
Create illusion of spectral sound.

Spectral sound is achieved by:
- delay of samples in left and right channel
- modification of amplitude according to distance between sound source and (left and right) ear

process() function is a function used by jack audio subsystem to process audio samples.
It contains application of spectral effects.

License: Creative Commons Zero
Author: Paweł Sabat
Website: sabat.e90.biz

See README for details.
*/
#include <stdio.h>
#include <jack/jack.h>
#include <SDL2/SDL.h>
#include <math.h>
#include <pthread.h>

int window_width = 640;
int window_height = 480;

jack_port_t *in;
jack_port_t *out_left, *out_right;
float sound_position[2] = {0.0, 0.0};
float man_position[2] = {0.0, 0.0};

/*
Difference of distances (right ear, sound source) and (left ear, sound source) cannot be bigger
than distance between ears (from triangle inequality), thus upper limit for number of samples that signal may
be delayed is distance_to_samples(0.2)=28, assuming that distance between ears is 0.2 meters.
*/
#define delay_buffer_max_length 28

float left_delay_buffer[delay_buffer_max_length];
float right_delay_buffer[delay_buffer_max_length];
int left_delay_samples = 0;
int right_delay_samples = 0;
int left_delay_samples_new = 0;
int right_delay_samples_new = 0;
float left_amplitude_scale = 1.0;
float right_amplitude_scale = 1.0;


// for passing values from main() to process() which are in separate threads
pthread_mutex_t shared_values_mutex;
int shared_left_delay = 0;
int shared_right_delay = 0;
float shared_left_amplitude = 1.0;
float shared_right_amplitude = 1.0;

// ============================================================================

void copy_samples_with_resize(float in[], int in_length, float out[], int out_length)
{
    if(out_length > 1)
    {
        // for stretch
        if(out_length > in_length)
        {
            // do a simple linear interpolation between two adjacent values
            // ..
            // I need last indexes as decimals so convert values to floats before doing calculations
            float in_last_pos = (float) in_length - 1;
            float out_last_pos = (float) out_length - 1;
            for(int i=0; i<out_length; ++i)
            {
                float in_position = i * in_last_pos / out_last_pos;
                // if position is after end of in array
                if(in_position >= in_last_pos)
                {
                    // use last point of in
                    out[i] = in[in_length - 1];
                }
                // if position is within limits of in array (between two points)
                else
                {
                    // interpolate between adjacent points of in
                    float integer, decimal;
                    decimal = modff(in_position, &integer);
                    float value = in[(int)floorf(in_position)] * (1.0 - decimal) + in[(int)ceilf(in_position)] * decimal;
                    out[i] = value;
                }
            }
        }
        // for shrink
        else
        {
            for(int i=0; i<out_length; ++i)
            {
                // just use samples as needed
                out[i] = in[i*(in_length-1)/(out_length-1)];
            }
        }
    }
    else if(out_length == 1)
    {
        out[0] = in[0];
    }
}

// ============================================================================

void copy_samples(float *input, float *output, int how_many)
{
    for(int i=0; i<how_many; ++i)
    {
        output[i] = input[i];
    }
}

// ============================================================================

void copy_samples_apply_delay(float *input, float *output, int nframes, float *delay_buffer,
    int old_delay, int new_delay)
{
    // no delay == just copy samples
    if(old_delay == 0 && new_delay == 0)
    {
        copy_samples(input, output, nframes);
    }
    // there is some constant delay
    else if(old_delay == new_delay)
    {
        // copy samples from delay buffer to output
        copy_samples(delay_buffer, output, old_delay);
        // copy (nframes - old_delay) samples from input to rest of output
        copy_samples(input, output + old_delay, nframes - old_delay);
        // copy last old_delay samples from input to delay buffer
        copy_samples(input + (nframes - old_delay), delay_buffer, old_delay);
    }
    // zero delay -> some delay
    else if(old_delay == 0 && new_delay > 0)
    {
        // copy samples from input to output without last new_delay samples
        // apply stretch to input samples to fill all output
        copy_samples_with_resize(input, nframes - new_delay, output, nframes);
        // copy last new_delay samples from input to delay buffer
        copy_samples(input + (nframes - new_delay), delay_buffer, new_delay);
    }
    // some delay -> zero delay
    else if(old_delay > 0 && new_delay == 0)
    {
        // copy old_delay samples from delay buffer to output
        copy_samples(delay_buffer, output, old_delay);
        // copy all samples from input to rest of output, apply shrink to input samples to fill rest of output
        copy_samples_with_resize(input, nframes, output + old_delay, nframes - old_delay);
    }
    // small delay -> BIG delay
    else if(old_delay < new_delay)
    {
        // copy all samples from delay buffer to output
        // delay_buffer has old_delay samples
        copy_samples(delay_buffer, output, old_delay);
        // copy samples from input to fill rest of output
        // rest of output == nframes - old_delay samples
        copy_samples(input, output + old_delay, nframes - old_delay);
        // copy samples from rest of input to delay buffer applying stretch
        // rest of input == old_delay samples
        copy_samples_with_resize(input + (nframes - old_delay), old_delay, delay_buffer, new_delay);
    }
    // BIG delay -> small delay
    else if(old_delay > new_delay)
    {
        // copy all samples from delay buffer to output
        // delay_buffer has old_delay samples
        copy_samples(delay_buffer, output, old_delay);
        // copy samples from input to fill rest of output
        // rest of output == nframes - old_delay samples
        copy_samples(input, output + old_delay, nframes - old_delay);
        // copy samples from rest of input to delay buffer applying shrink
        // rest of input == old_delay samples
        copy_samples_with_resize(input + (nframes - old_delay), old_delay, delay_buffer, new_delay);
    }
}


// ============================================================================
/*
This function will be run in separate thread by jack to process audio samples.
Samples are fed in chunks eg. 1024 samples per call.
nframes param tells how many samples are to be processed.

It processes samples in real time, so it shall not make any prints, file reads etc.
*/
int process(jack_nframes_t nframes, void *arg)
{
    (void)arg; // ignore unused parameter
    float *input_buffer = (float*)jack_port_get_buffer(in, nframes);
    float *output_left_buffer = (float*)jack_port_get_buffer(out_left, nframes);
    float *output_right_buffer = (float*)jack_port_get_buffer(out_right, nframes);

    // update values as often as possible, but do not wait for mutex if busy
    if(pthread_mutex_trylock(&shared_values_mutex) == 0)
    {
        left_delay_samples_new = shared_left_delay;
        right_delay_samples_new = shared_right_delay;
        left_amplitude_scale = shared_left_amplitude;
        right_amplitude_scale = shared_right_amplitude;

        pthread_mutex_unlock(&shared_values_mutex);
    }

    // copy samples from input to left channel applying delay
    copy_samples_apply_delay(input_buffer, output_left_buffer, nframes, left_delay_buffer,
        left_delay_samples, left_delay_samples_new);
    // copy samples from input to right channel applying delay
    copy_samples_apply_delay(input_buffer, output_right_buffer, nframes, right_delay_buffer,
        right_delay_samples, right_delay_samples_new);

    // make old delay values a current one
    left_delay_samples = left_delay_samples_new;
    right_delay_samples = right_delay_samples_new;

    // apply amplitude adjustment to output samples
    for(unsigned int i=0; i<nframes; ++i)
    {
        output_left_buffer[i] *= left_amplitude_scale;
        output_right_buffer[i] *= right_amplitude_scale;
    }

    return 0;
}


// ============================================================================

// compute by how many samples signal is delayed
// it it has additional distance (in meters) to travel
// assuming 48kHz sampling, sound speed in air = 343 m/s
int distance_to_samples(float distance)
{
    float distance_sec = distance / 343.0;
    float single_sample_duration = 1.0 / 48000; // in seconds
    int sample_count = roundf(distance_sec / single_sample_duration);
    return sample_count;
}

// ============================================================================

// handle all the updates required when sound changes position like changing sound delay and loudness
void sound_position_changed()
{
    /*
    compute distance of left and right ear of the man from the sound source
    assuming distance between ears is 20cm in straight line, then
    left ear is -0.1 from man x position, and right ear is +0.1 from man x position
    assuming no rotation (change of perspective) is possible
    */
    float left_ear_distance = hypotf(man_position[0] - 0.1 - sound_position[0], man_position[1] - sound_position[1]);
    float right_ear_distance = hypotf(man_position[0] + 0.1 - sound_position[0], man_position[1] - sound_position[1]);

    // compute difference of these distances
    float distance_diff = fabsf(left_ear_distance - right_ear_distance);
    int delayed_sample_count = distance_to_samples(distance_diff);

    // set shared delay values

    // lock mutex for shared values
    // ..
    pthread_mutex_lock(&shared_values_mutex);

    // set amplitude scaling factors based on distance
    shared_left_amplitude = powf(1.0/2.0, left_ear_distance/100.0);
    shared_right_amplitude = powf(1.0/2.0, right_ear_distance/100.0);

    // see which ear shall have delayed signal
    if(left_ear_distance >= right_ear_distance)
    {
        // left ear shall have delayed signal
        shared_left_delay = delayed_sample_count;
        // right shall not
        shared_right_delay = 0;
    }
    else
    {
        // right ear shall have delayed signal
        shared_right_delay = delayed_sample_count;
        // left shall not
        shared_left_delay = 0;
    }

    // ..
    // unlock mutex for shared values
    pthread_mutex_unlock(&shared_values_mutex);

    printf("amplitude scale left=%f right=%f\n", shared_left_amplitude, shared_right_amplitude);
    printf("delay samples left=%d right=%d\n", shared_left_delay, shared_right_delay);
}

// ============================================================================

// center of window is position (x=0.0, y=0.0)
void position_to_window_coordinates(float x, float y, int *wx, int *wy)
{
    *wx = (int)(window_width/2 + x);
    *wy = (int)(window_height/2 + y);
}

// ============================================================================

void window_to_position_coordinates(int wx, int wy, float *x, float *y)
{
    *x = wx - window_width/2;
    *y = wy - window_height/2;
}

// ============================================================================

void draw_scene(SDL_Renderer *renderer)
{
    int sound_x, sound_y;
    int man_x, man_y;
    int point_width = 10, point_height = 10;
    SDL_Rect rect;
    rect.w = point_width;
    rect.h = point_height;

    // clear window
    SDL_SetRenderDrawColor(renderer, 0, 0, 0, 255);
    SDL_RenderClear(renderer);

    // get window coordinates of man and sound source
    position_to_window_coordinates(sound_position[0], sound_position[1], &sound_x, &sound_y);
    position_to_window_coordinates(man_position[0], man_position[1], &man_x, &man_y);

    // draw line connecting sound source and man
    SDL_SetRenderDrawColor(renderer, 100, 100, 0, 255);
    SDL_RenderDrawLine(renderer, sound_x, sound_y, man_x, man_y);

    // draw sound source
    SDL_SetRenderDrawColor(renderer, 200, 200, 0, 255);
    rect.x = sound_x - point_width/2;
    rect.y = sound_y - point_width/2;
    SDL_RenderFillRect(renderer, &rect);

    // draw man
    SDL_SetRenderDrawColor(renderer, 250, 250, 200, 255);
    rect.x = man_x - point_width/2;
    rect.y = man_y - point_width/2;
    SDL_RenderFillRect(renderer, &rect);

    // show current scene
    SDL_RenderPresent(renderer);
}

// ============================================================================

int init_jack_subsystem(jack_client_t **jack_client_var)
{
    // connect to jack server
    jack_client_t *client = jack_client_open("spectral", JackNoStartServer, NULL);
    if(client == NULL)
    {
        fprintf(stderr, "Connecting to jack failed\n");
        return -1;
    }

    // create input port
    in = jack_port_register(client, // client for which to create port
        "in",                    // name of a port
        JACK_DEFAULT_AUDIO_TYPE, // port type telling what kind of data it handles
        JackPortIsInput,         // port flags, most important is telling if this port is input or output
        0);                      // not used, parameter ignored for builtin type ports
    if(in == NULL)
    {
        fprintf(stderr, "Creating input port failed\n");
        jack_client_close(client);
        return -1;
    }

    // create output ports
    out_left = jack_port_register(client, "out-left", JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);
    out_right = jack_port_register(client, "out-right", JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);
    if(out_left == NULL || out_right == NULL)
    {
        fprintf(stderr, "Creating output ports failed\n");
        // ports are unregistered automatically on client close, so there is no need to unregister in port
        jack_client_close(client);
        return -1;
    }

    // register process function
    jack_set_process_callback(client, process, NULL);

    // initialize mutex for passing values between main() and process()
    pthread_mutex_init(&shared_values_mutex, NULL);

    // start processing audio
    jack_activate(client);

    // make automatic connections between ports
    jack_connect(client, "spectral:out-left", "system:playback_1");
    jack_connect(client, "spectral:out-right", "system:playback_2");

    printf("jack connected and running\n");

    *jack_client_var = client;
    return 0;
}

// ============================================================================

void deinit_jack_subsystem(jack_client_t *client)
{
    jack_client_close(client);
    pthread_mutex_destroy(&shared_values_mutex);
}

// ============================================================================

int init_sdl_subsystem(SDL_Window **window_var, SDL_Renderer **renderer_var)
{
    if(SDL_Init(SDL_INIT_VIDEO) != 0)
    {
        fprintf(stderr, "Error initializing SDL2.\n%s\n", SDL_GetError());
        return -1;
    }

    SDL_Window *window = SDL_CreateWindow("Hello World!", 100, 100, window_width, window_height, SDL_WINDOW_SHOWN);
    if (window == NULL)
    {
        fprintf(stderr, "Create window error: %s\n", SDL_GetError());
        SDL_Quit();
        return -1;
    }

    SDL_Renderer *renderer = SDL_CreateRenderer(window, -1, 0);
    if(renderer == NULL)
    {
        fprintf(stderr, "Create window renderer error: %s\n", SDL_GetError());
        SDL_DestroyWindow(window);
        SDL_Quit();
        return -1;
    }

    printf("window and events initialized\n");

    *window_var = window;
    *renderer_var = renderer;
    return 0;
}

// ============================================================================

void deinit_sdl_subsystem(SDL_Window *window, SDL_Renderer *renderer)
{
    SDL_DestroyRenderer(renderer);
    SDL_DestroyWindow(window);
    SDL_Quit();
}

// ============================================================================

int main(void)
{
    jack_client_t *client;
    SDL_Window *window;
    SDL_Renderer *renderer;
    int r;

    r = init_jack_subsystem(&client);
    if(r != 0)
    {
        fprintf(stderr, "Could not initialize jack audio subsystem.\n");
        return 1;
    }

    r = init_sdl_subsystem(&window, &renderer);
    if(r != 0)
    {
        fprintf(stderr, "Could not initialize sdl window and events subsystem.\n");
        deinit_jack_subsystem(client);
        return 1;
    }

    // program main loop
    while(1)
    {
        SDL_Event event;
        while(SDL_PollEvent(&event))
        {
            if(event.type == SDL_QUIT)
            {
                printf("Exiting program\n");
                deinit_sdl_subsystem(window, renderer);
                deinit_jack_subsystem(client);
                return 0;
            }
            else if(event.type == SDL_WINDOWEVENT)
            {
                draw_scene(renderer);
            }
            else if(event.type == SDL_MOUSEMOTION)
            {
                if((event.motion.state & SDL_BUTTON_LMASK) != 0)
                {
                    int wx = event.motion.x;
                    int wy = event.motion.y;
                    window_to_position_coordinates(wx, wy, &sound_position[0], &sound_position[1]);
                    sound_position_changed();
                    draw_scene(renderer);
                }
            }
            else if(event.type == SDL_MOUSEBUTTONDOWN)
            {
                int wx = event.button.x;
                int wy = event.button.y;
                window_to_position_coordinates(wx, wy, &sound_position[0], &sound_position[1]);
                sound_position_changed();
                draw_scene(renderer);
            }
        }

        SDL_Delay(20); // sleep for some time in ms
    }

    return 0;
}
